from fastapi_controllers.controllers import Controller
from fastapi_controllers.routing import delete, get, head, options, patch, post, put, trace, websocket

__all__ = [
    "Controller",
    "delete",
    "get",
    "head",
    "options",
    "patch",
    "post",
    "put",
    "trace",
    "websocket",
]



from typing import Any, Callable, Dict, Tuple

from fastapi_controllers.definitions import Route, RouteData, RouteDefinition
from fastapi_controllers.helpers import _validate_against_signature


class _RouteDecorator:
    _route_definition: RouteDefinition

    def __init_subclass__(cls, route_definition: RouteDefinition) -> None:
        super().__init_subclass__()
        cls._route_definition = route_definition

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.route_args = args
        self.route_kwargs = kwargs
        _validate_against_signature(self._route_definition.binds, args=args, kwargs=kwargs)

    def __call__(self, func: Callable[..., Any]) -> Callable[..., Any]:
        func.__route_data__ = RouteData(  # type: ignore
            route_definition=self._route_definition,
            route_args=self.route_args,
            route_kwargs=self.route_kwargs,
        )
        return func

    @property
    def route_args(self) -> Tuple[Any, ...]:
        return self._route_args

    @route_args.setter
    def route_args(self, value: Tuple[Any, ...]) -> None:
        self._route_args = value

    @property
    def route_kwargs(self) -> Dict[str, Any]:
        return self._route_kwargs

    @route_kwargs.setter
    def route_kwargs(self, value: Dict[str, Any]) -> None:
        self._route_kwargs = value


class delete(_RouteDecorator, route_definition=Route.delete):
    ...


class get(_RouteDecorator, route_definition=Route.get):
    ...


class head(_RouteDecorator, route_definition=Route.head):
    ...


class options(_RouteDecorator, route_definition=Route.options):
    ...


class patch(_RouteDecorator, route_definition=Route.patch):
    ...


class post(_RouteDecorator, route_definition=Route.post):
    ...


class put(_RouteDecorator, route_definition=Route.put):
    ...


class trace(_RouteDecorator, route_definition=Route.trace):
    ...


class websocket(_RouteDecorator, route_definition=Route.websocket):
    ...



from unittest.mock import MagicMock

import pytest
from fastapi import APIRouter
from pytest_mock import MockerFixture

from fastapi_controllers.controllers import Controller
from fastapi_controllers.routing import get, websocket


@pytest.fixture(autouse=True)
def validator(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("fastapi_controllers.controllers._validate_against_signature")


def describe_Controller() -> None:
    def it_sets_router_params() -> None:
        class FakeController(Controller):
            prefix = "/test"
            dependencies = ["TEST"]  # type: ignore
            tags = ["TEST"]
            __router_params__ = {
                "prefix": "/override",
                "tags": ["override"],
                "deprecated": True,
            }

        assert FakeController.__router_params__ == {
            "prefix": "/override",
            "dependencies": ["TEST"],
            "tags": ["override"],
            "deprecated": True,
        }

    def it_validates_apirouter_parameters(mocker: MockerFixture, validator: MagicMock) -> None:

        mocked_proxy = mocker.patch("weakref.proxy", return_value="FAKE_PROXY")

        class _(Controller):
            ...

        mocked_proxy.assert_called_once_with(APIRouter.__init__)
        validator.assert_called_once_with(
            "FAKE_PROXY",
            kwargs={
                "prefix": "",
                "dependencies": None,
                "tags": None,
            },
        )

    def describe_create_router() -> None:
        def it_creates_an_apirouter() -> None:
            class FakeController(Controller):
                prefix = "/test"

                @get("/get", deprecated=True)
                def fake_method(self) -> None:
                    ...

            result = FakeController.create_router()
            assert isinstance(result, APIRouter)

        def it_replaces_method_signatures(mocker: MockerFixture) -> None:
            replace = mocker.patch("fastapi_controllers.controllers._replace_signature")

            class FakeController(Controller):
                prefix = "/test"

                @get("/get", deprecated=True)
                def fake_method(self) -> None:
                    ...

            FakeController.create_router()
            replace.assert_called_once_with(FakeController, FakeController.fake_method)

        def it_configures_the_router_and_routes(mocker: MockerFixture) -> None:
            apirouter = mocker.patch("fastapi_controllers.controllers.APIRouter")

            class FakeController(Controller):
                prefix = "/test"

                @get("/get", deprecated=True)
                def fake_method(self) -> None:
                    ...

                @websocket("/ws")
                def fake_ws(self) -> None:
                    ...

            FakeController.create_router()
            apirouter.assert_called_once_with(prefix="/test", dependencies=None, tags=None)
            apirouter.return_value.add_api_route.assert_called_once_with(
                "/get",
                FakeController.fake_method,
                deprecated=True,
                methods=["GET"],
            )
            apirouter.return_value.add_api_websocket_route.assert_called_once_with(
                "/ws",
                FakeController.fake_ws,
            )



import inspect
import weakref
from typing import Any

import pytest
from fastapi import params

from fastapi_controllers.helpers import _replace_signature, _validate_against_signature


class Fake:
    def fake_method(self, positional: Any, *, keyword: "str") -> None:
        ...


def describe_validate_against_signature() -> None:
    def it_validates_method_parameters_against_the_desired_signature() -> None:
        _validate_against_signature(
            weakref.proxy(Fake.fake_method),
            args=("test",),
            kwargs={"keyword": "TEST"},
        )

    def it_raises_an_error_on_missing_positional_args() -> None:
        with pytest.raises(TypeError):
            _validate_against_signature(
                weakref.proxy(Fake.fake_method),
                args=tuple(),
                kwargs={"keyword": "TEST"},
            )

    def it_raises_an_error_on_missing_keyword_args() -> None:
        with pytest.raises(TypeError):
            _validate_against_signature(
                weakref.proxy(Fake.fake_method),
                args=("test",),
                kwargs={},
            )

    def it_raises_an_error_on_additional_keyword_args() -> None:
        with pytest.raises(TypeError):
            _validate_against_signature(
                weakref.proxy(Fake.fake_method),
                args=("test",),
                kwargs={"keyword": "TEST", "additional": "TEST"},
            )

    def it_raises_an_error_on_additional_positional_args() -> None:
        with pytest.raises(TypeError):
            _validate_against_signature(
                weakref.proxy(Fake.fake_method),
                args=("test", "test"),
                kwargs={},
            )


def describe_replace_signature() -> None:
    def it_replaces_the_signature_of_an_instance_method() -> None:
        class Test:
            def test(self, positional: Any, *, keyword: str) -> None:
                ...

        _replace_signature(Test, Test.test)
        replaced_params = list(inspect.signature(Test.test).parameters.values())
        assert replaced_params[0].name == "self"
        assert replaced_params[0].kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
        assert isinstance(replaced_params[0].default, params.Depends)
        assert replaced_params[0].default.dependency == Test
        assert replaced_params[1].name == "positional"
        assert replaced_params[1].kind == inspect.Parameter.KEYWORD_ONLY
        assert replaced_params[2].name == "keyword"
        assert replaced_params[2].kind == inspect.Parameter.KEYWORD_ONLY



import inspect
import weakref
from typing import Any, Callable, Dict, Optional, Tuple, Type

from fastapi import Depends


def _validate_against_signature(
    method: weakref.CallableProxyType,
    args: Optional[Tuple[Any, ...]] = None,
    kwargs: Optional[Dict[str, Any]] = None,
) -> None:
    """
    Validate method parameters against those of the corresponding APIRouter method.

    Args:
        method_name: The name of an APIRouter method to validate against.
        args: The positional arguments of the method.
        kwargs: The keyword arguments of the method.
    """
    target_sig = inspect.signature(method)
    valid_sig = target_sig.replace(parameters=list(target_sig.parameters.values())[1:])
    valid_sig.bind(*(args or tuple()), **(kwargs or {}))


def _replace_signature(cls: Type, func: Callable[..., Any]) -> None:
    orig_sig = inspect.signature(func)
    new_params = [
        param.replace(
            kind=inspect.Parameter.KEYWORD_ONLY,
        )
        if param.name != "self"
        else param.replace(default=Depends(cls))
        for param in list(orig_sig.parameters.values())
    ]
    func.__signature__ = orig_sig.replace(parameters=new_params)  # type: ignore



import weakref
from typing import Type
from unittest.mock import MagicMock

import pytest
from pytest_mock import MockerFixture

from fastapi_controllers.definitions import HTTPRequestMethod, RouteData, RouteDefinition
from fastapi_controllers.routing import _RouteDecorator, delete, get, head, options, patch, post, put, trace

HTTP_DECO_DEFINITIONS = {
    delete: HTTPRequestMethod.DELETE,
    get: HTTPRequestMethod.GET,
    head: HTTPRequestMethod.HEAD,
    options: HTTPRequestMethod.OPTIONS,
    patch: HTTPRequestMethod.PATCH,
    post: HTTPRequestMethod.POST,
    put: HTTPRequestMethod.PUT,
    trace: HTTPRequestMethod.TRACE,
}


class FakeDefinition(RouteDefinition):
    ...


def original_method() -> None:
    ...


def fake_method() -> None:
    ...


binds = weakref.proxy(original_method)


class fake(_RouteDecorator, route_definition=FakeDefinition(binds=binds)):
    ...


@pytest.fixture
def validator(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("fastapi_controllers.routing._validate_against_signature")


def describe_RouteDecorator() -> None:
    def it_validates_the_parameters_of_the_decorated_method(validator: MagicMock) -> None:
        fake("/test", keyword="TEST")(fake_method)
        validator.assert_called_once_with(
            fake._route_definition.binds,
            args=("/test",),
            kwargs={
                "keyword": "TEST",
            },
        )

    def it_adds_route_data_to_the_decorated_method(validator: MagicMock) -> None:
        wrapped = fake("/test", keyword="TEST")(fake_method)
        assert isinstance(wrapped.__route_data__, RouteData)  # type: ignore
        assert isinstance(wrapped.__route_data__.route_definition, FakeDefinition)  # type: ignore
        assert wrapped.__route_data__.route_definition.binds == binds  # type: ignore
        assert wrapped.__route_data__.route_args == ("/test",)  # type: ignore
        assert wrapped.__route_data__.route_kwargs == {"keyword": "TEST"}  # type: ignore


def describe_decorators() -> None:
    @pytest.mark.parametrize("decorator", HTTP_DECO_DEFINITIONS.keys())
    def it_is_a_subclass_of_route_decorator(decorator: Type[_RouteDecorator]) -> None:
        assert issubclass(decorator, _RouteDecorator)

    @pytest.mark.parametrize("decorator,method", HTTP_DECO_DEFINITIONS.items())
    def it_handles_the_proper_http_request_method(
        decorator: Type[_RouteDecorator],
        method: HTTPRequestMethod,
    ) -> None:
        assert decorator._route_definition.request_method == method  # type: ignore



import pytest
from fastapi import status
from fastapi.testclient import TestClient
from starlette.websockets import WebSocketDisconnect

from fastapi_controllers.definitions import HTTPRequestMethod


def describe_test_controller_async() -> None:
    @pytest.mark.parametrize("http_request_method", HTTPRequestMethod.__members__.values())
    def it_responds_to_http_methods(
        async_test_client: TestClient,
        http_request_method: HTTPRequestMethod,
    ) -> None:
        request_func = getattr(async_test_client, http_request_method.lower(), None)
        if request_func:
            response = request_func("/test-async")
            assert response.status_code == status.HTTP_200_OK

    def it_supports_websockets(async_test_client: TestClient) -> None:
        with pytest.raises(WebSocketDisconnect):
            with async_test_client.websocket_connect("/ws") as websocket:
                data = websocket.receive_json()
                assert data == {"msg": "Hello WebSocket"}

    def it_resolves_async_dependencies(async_test_client: TestClient) -> None:
        response = async_test_client.get("/test-async")
        assert response.status_code == status.HTTP_200_OK
        assert response.text == "ASYNC TEST"



import pytest
from fastapi import status
from fastapi.testclient import TestClient

from fastapi_controllers.definitions import HTTPRequestMethod


def describe_test_controller_sync() -> None:
    @pytest.mark.parametrize("http_request_method", HTTPRequestMethod.__members__.values())
    def it_responds_to_http_methods(
        sync_test_client: TestClient,
        http_request_method: HTTPRequestMethod,
    ) -> None:
        request_func = getattr(sync_test_client, http_request_method.lower(), None)
        if request_func:
            response = request_func("/test-sync")
            assert response.status_code == status.HTTP_200_OK

    def it_resolves_sync_dependencies(sync_test_client: TestClient) -> None:
        response = sync_test_client.get("/test-sync")
        assert response.status_code == status.HTTP_200_OK
        assert response.text == "SYNC TEST"



import asyncio
import time

import pytest
from fastapi import Depends, FastAPI, Response, status
from fastapi.testclient import TestClient
from fastapi.websockets import WebSocket

from fastapi_controllers import Controller, delete, get, head, options, patch, post, put, trace, websocket


def sync_dependency() -> str:
    # DO SOME FAKE SYNC STUFF
    time.sleep(0.01)
    return "SYNC TEST"


class SyncTestController(Controller):
    prefix = "/test-sync"

    def __init__(self, message: str = Depends(sync_dependency)) -> None:  # noqa: B008
        self.message = message

    @delete("", response_class=Response)
    def test_delete(self) -> Response:
        return Response(status_code=status.HTTP_200_OK)

    @get("", response_class=Response)
    def test_get(self) -> Response:
        return Response(content=self.message, status_code=status.HTTP_200_OK)

    @head("", response_class=Response)
    def test_head(self) -> Response:
        return Response(status_code=status.HTTP_200_OK)

    @options("", response_class=Response)
    def test_options(self) -> Response:
        return Response(status_code=status.HTTP_200_OK)

    @patch("", response_class=Response)
    def test_patch(self) -> Response:
        return Response(status_code=status.HTTP_200_OK)

    @post("", response_class=Response)
    def test_post(self) -> Response:
        return Response(status_code=status.HTTP_200_OK)

    @put("", response_class=Response)
    def test_put(self) -> Response:
        return Response(status_code=status.HTTP_200_OK)

    @trace("", response_class=Response)
    def test_trace(self) -> Response:
        return Response(status_code=status.HTTP_200_OK)


async def async_dependency() -> str:
    # DO SOME FAKE ASYNC STUFF
    await asyncio.sleep(0.01)
    return "ASYNC TEST"


class AsyncTestController(Controller):
    prefix = "/test-async"

    def __init__(self, message: str = Depends(async_dependency)) -> None:  # noqa: B008
        self.message = message

    @delete("", response_class=Response)
    async def test_delete(self) -> Response:
        return Response(status_code=status.HTTP_200_OK)

    @get("", response_class=Response)
    async def test_get(self) -> Response:
        return Response(content=self.message, status_code=status.HTTP_200_OK)

    @head("", response_class=Response)
    async def test_head(self) -> Response:
        return Response(status_code=status.HTTP_200_OK)

    @options("", response_class=Response)
    async def test_options(self) -> Response:
        return Response(status_code=status.HTTP_200_OK)

    @patch("", response_class=Response)
    async def test_patch(self) -> Response:
        return Response(status_code=status.HTTP_200_OK)

    @post("", response_class=Response)
    async def test_post(self) -> Response:
        return Response(status_code=status.HTTP_200_OK)

    @put("", response_class=Response)
    async def test_put(self) -> Response:
        return Response(status_code=status.HTTP_200_OK)

    @trace("", response_class=Response)
    async def test_trace(self) -> Response:
        return Response(status_code=status.HTTP_200_OK)

    @websocket("/ws")
    async def websocket(websocket: WebSocket) -> None:  # type: ignore
        await websocket.accept()
        await websocket.send_json({"msg": "Hello WebSocket"})
        await websocket.close()


@pytest.fixture
def sync_test_client() -> TestClient:
    app = FastAPI()
    app.include_router(SyncTestController.create_router())
    return TestClient(app)


@pytest.fixture
def async_test_client() -> TestClient:
    app = FastAPI()
    app.include_router(AsyncTestController.create_router())
    return TestClient(app)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastapi_controllers', 'tests', 'tests.functional', 'tests.unit']

package_data = \
{'': ['*']}

install_requires = \
['fastapi']

setup_kwargs = {
    'name': 'fastapi-controllers',
    'version': '0.2.1',
    'description': 'Simple Controller implementation for FastAPI',
    'long_description': '<p>\n    <h1 align="center">fastapi-controllers</h1>\n    <p align="center">\n        A simple solution for organizing your FastAPI endpoints\n    </p>\n    <p align="center">\n    <img src="https://badgen.net/pypi/v/fastapi-controllers"/>\n    <img src="https://badgen.net/pypi/license/fastapi-controllers"/>\n    <img src="https://badgen.net/pypi/python/fastapi-controllers"/>\n    <img src="https://github.com/goraje/fastapi-controllers/actions/workflows/test.yml/badge.svg">\n    </p>\n</p>\n\n## Organize your API endpoints\n\n`fastapi-controllers` offers a simple solution for organizing your API endpoints by means of a `Controller` class embracing the concept of class-based views.\n\n## Features\n\n- class-based approach to organizing FastAPI endpoints\n- class-scoped definition of APIRouter parameters\n- instance-scoped definition of FastAPI dependencies\n- it integrates seamlessly with the FastAPI framework\n- works with both sync and async endpoints\n\n## Installation\n\n```sh\npip install fastapi-controllers\n```\n\n## Minimal working example\n\n```python\nimport uvicorn\nfrom fastapi import FastAPI, Response, status\nfrom fastapi.websockets import WebSocket\n\nfrom fastapi_controllers import Controller, get, websocket\n\n\nclass ExampleController(Controller):\n    @get("/example", response_class=Response)\n    async def get_example(self) -> Response:\n        return Response(status_code=status.HTTP_200_OK)\n\n    @websocket("/ws")\n    async def ws_example(websocket: WebSocket) -> None:\n        await websocket.accept()\n        while True:\n            data = await websocket.receive_text()\n            await websocket.send_text(f"Received: {data}")\n\n\nif __name__ == "__main__":\n    app = FastAPI()\n    app.include_router(ExampleController.create_router())\n    uvicorn.run(app)\n```\n\nFastAPI\'s `APIRouter` is created and populated with API routes by the `Controller.create_router` method and can be incorporated into the application in the usual way via `app.include_router`.\n\n## Seamless integration\n\nThe router-related parameters as well as those of HTTP request-specific and websocket decorators are expected to be the same as those used by `fastapi.APIRouter`, `fastapi.APIRouter.<request_method>` and `fastapi.APIRouter.websocket`. Validation of the provided parameters is performed during initialization via the `inspect` module. This ensures compatibility with the FastAPI framework and prevents the introduction of a new, unnecessary naming convention.\n\n### Available decorators\n\n```python\nfrom fastapi_controllers import delete, get, head, options, patch, post, put, trace, websocket\n```\n\n## Use class variables to customize your APIRouter\n\nClass variables can be used to set the commonly used APIRouter parameters: `prefix`, `dependencies` and `tags`.\n\n```python\nimport uvicorn\nfrom fastapi import Depends, FastAPI, Response, status\nfrom fastapi.security import HTTPBasic, HTTPBasicCredentials\nfrom pydantic import BaseModel\n\nfrom fastapi_controllers import Controller, get, post\n\nsecurity = HTTPBasic()\n\n\nasync def authorized_user(credentials: HTTPBasicCredentials = Depends(security)) -> None:\n    ...\n\n\nclass ExampleRequest(BaseModel):\n    name: str\n\n\nclass ExampleResponse(BaseModel):\n    message: str\n\n\nclass ExampleController(Controller):\n    prefix = "/example"\n    tags = ["example"]\n    dependencies = [Depends(authorized_user)]\n\n    @get("", response_class=Response)\n    async def get_example(self) -> Response:\n        return Response(status_code=status.HTTP_200_OK)\n\n    @post("", response_model=ExampleResponse)\n    async def post_example(self, data: ExampleRequest) -> ExampleResponse:\n        return ExampleResponse(message=f"Hello, {data.name}!")\n\n\nif __name__ == "__main__":\n    app = FastAPI()\n    app.include_router(ExampleController.create_router())\n    uvicorn.run(app)\n```\n### Additional APIRouter parameters\nAdditional APIRouter parameters can be provided via the `__router_params__` class variable in form of a mapping.\n\n```python\nimport uvicorn\nfrom fastapi import FastAPI, Response, status\n\nfrom fastapi_controllers import Controller, get\n\n\nclass ExampleController(Controller):\n    prefix = "/example"\n    tags = ["example"]\n    __router_params__ = {"deprecated": True}\n\n    @get("", response_class=Response)\n    async def get_example(self) -> Response:\n        return Response(status_code=status.HTTP_200_OK)\n\n\nif __name__ == "__main__":\n    app = FastAPI()\n    app.include_router(ExampleController.create_router())\n    uvicorn.run(app)\n```\n> :warning: **Important**: Beware of assigning values to the same parameter twice (directly on class-level and through `__router_params__`). The values stored in `__router_params__` have precedence and will override your other settings if a name conflict arises. E.g. the following `Controller` would create an `APIRouter` with `prefix=/override`, `tags=["override"]` and `dependencies=[Depends(override)]`\n\n```python\nfrom fastapi import Depends\n\nfrom fastapi_controllers import Controller\n\n\nclass ExampleController(Controller):\n    prefix = "/example"\n    tags = ["example"]\n    dependencies = [Depends(example)]\n    __router_params__ = {\n        "prefix": "/override",\n        "tags": ["override"],\n        "dependencies": [Depends(override)],\n    }\n```\n\n\n## Instance-scoped dependencies\n\nInstance-scoped attributes can be defined in the `__init__` method of the `Controller` and offer an easy way to access common dependencies for all endpoints.\n\n```python\nimport json\n\nimport uvicorn\nfrom fastapi import Depends, FastAPI, Response, status\n\nfrom fastapi_controllers import Controller, get\n\n\nclass DbSession:\n    @property\n    def status(self) -> str:\n        return "CONNECTED"\n\n\nasync def get_db_session() -> DbSession:\n    return DbSession()\n\n\nclass ExampleController(Controller):\n    prefix = "/example"\n\n    def __init__(self, session: DbSession = Depends(get_db_session)) -> None:\n        self.session = session\n\n    @get("", response_class=Response)\n    async def get_status(self) -> Response:\n        return Response(\n            content=json.dumps({"status": f"{self.session.status}"}),\n            status_code=status.HTTP_200_OK,\n            media_type="application/json",\n        )\n\n\nif __name__ == "__main__":\n    app = FastAPI()\n    app.include_router(ExampleController.create_router())\n    uvicorn.run(app)\n```\n',
    'author': 'Jerzy Góra',
    'author_email': 'j.gora89@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/goraje/fastapi-controllers',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import weakref
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Tuple

from fastapi import APIRouter


class HTTPRequestMethod(str, Enum):
    DELETE = "DELETE"
    GET = "GET"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    PATCH = "PATCH"
    POST = "POST"
    PUT = "PUT"
    TRACE = "TRACE"


class RouteDefinition:
    def __init__(self, *, binds: weakref.CallableProxyType) -> None:
        self.binds = binds


class HTTPRouteDefinition(RouteDefinition):
    def __init__(self, *, binds: weakref.CallableProxyType, request_method: HTTPRequestMethod) -> None:
        super().__init__(binds=binds)
        self.request_method = request_method


class WebsocketRouteDefinition(RouteDefinition):
    ...


@dataclass
class RouteData:
    route_definition: RouteDefinition
    route_args: Tuple[Any, ...]
    route_kwargs: Dict[str, Any]


@dataclass
class Route:
    delete = HTTPRouteDefinition(binds=weakref.proxy(APIRouter.delete), request_method=HTTPRequestMethod.DELETE)
    get = HTTPRouteDefinition(binds=weakref.proxy(APIRouter.get), request_method=HTTPRequestMethod.GET)
    head = HTTPRouteDefinition(binds=weakref.proxy(APIRouter.head), request_method=HTTPRequestMethod.HEAD)
    options = HTTPRouteDefinition(binds=weakref.proxy(APIRouter.options), request_method=HTTPRequestMethod.OPTIONS)
    patch = HTTPRouteDefinition(binds=weakref.proxy(APIRouter.patch), request_method=HTTPRequestMethod.PATCH)
    post = HTTPRouteDefinition(binds=weakref.proxy(APIRouter.post), request_method=HTTPRequestMethod.POST)
    put = HTTPRouteDefinition(binds=weakref.proxy(APIRouter.put), request_method=HTTPRequestMethod.PUT)
    trace = HTTPRouteDefinition(binds=weakref.proxy(APIRouter.trace), request_method=HTTPRequestMethod.TRACE)
    websocket = WebsocketRouteDefinition(binds=weakref.proxy(APIRouter.websocket))



import inspect
import weakref
from enum import Enum
from typing import Any, Dict, List, Optional, Sequence, Union

from fastapi import APIRouter, params

from fastapi_controllers.definitions import HTTPRouteDefinition, RouteData, WebsocketRouteDefinition
from fastapi_controllers.helpers import _replace_signature, _validate_against_signature


class Controller:
    prefix: str = ""
    dependencies: Optional[Sequence[params.Depends]] = None
    tags: Optional[List[Union[str, Enum]]] = None
    __router_params__: Optional[Dict[str, Any]] = None

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        cls.__router_params__ = cls.__router_params__ or {}
        for param in ["prefix", "dependencies", "tags"]:
            if not cls.__router_params__.get(param):
                cls.__router_params__[param] = getattr(cls, param)
        _validate_against_signature(weakref.proxy(APIRouter.__init__), kwargs=cls.__router_params__)

    @classmethod
    def create_router(cls) -> APIRouter:
        """
        Create a new APIRouter instance and populate the APIRoutes.

        Returns:
            APIRouter: An APIRouter instance.
        """
        router = APIRouter(**cls.__router_params__)  # type: ignore
        for _, func in inspect.getmembers(cls, predicate=inspect.isfunction):
            _replace_signature(cls, func)
            route_data: Optional[RouteData] = getattr(func, "__route_data__", None)
            if route_data:
                if isinstance(route_data.route_definition, HTTPRouteDefinition):
                    router.add_api_route(
                        route_data.route_args[0],
                        func,
                        *route_data.route_args[1:],
                        methods=[route_data.route_definition.request_method],
                        **route_data.route_kwargs,
                    )
                if isinstance(route_data.route_definition, WebsocketRouteDefinition):
                    router.add_api_websocket_route(
                        route_data.route_args[0],
                        func,
                        *route_data.route_args[1:],
                        **route_data.route_kwargs,
                    )
        return router
